#include "VASupport.h"

//#import <Foundation/Foundation.h>

/*
 * All C functions provided in whole or in part via sample programs by ValiCert
 */

/*
 * Copyright 1999 ValiCert Inc. All Rights Reserved.
 * $Id: vtksample.c,v 1.6 2000/03/02 03:56:43 stefank Exp $
 * ValiCert Validator Toolkit Sample Program.
 * This file demonstrates performing common operations with the
 * ValidCert Validator Toolkit.
 */



/*
 * showStatus
 * Displays textual representation of the passed in validation
 * status code.
 */
static void showStatus(const char*label, Vtk_uint32 status);

static void assignValidationDetails(const Vtk_Ctxt *ctxt, const Vtk_ValHdl *hdl, NSMutableDictionary *res);

NSString *statusMessageForStatus(Vtk_uint32 status) {
    if(status & VTK_STATUS_OK){
        return @"OK";
    }else if (status & VTK_STATUS_REVOKED){
        return @"REVOKED";
    }else if (status & VTK_STATUS_UNKNOWN){
        return @"UNKNOWN";
    }else if (status & VTK_STATUS_CRL_EXPIRED){
        return @"CRL_EXPIRED";
    }else if (status & VTK_STATUS_CRL_NOT_YET_VALID){
        return @"CRL_NOT_YET_VALID";
    }else if (status & VTK_STATUS_RESP_EXPIRED){
        return @"STATUS_RESP_EXPIRED";
    }else if (status & VTK_STATUS_RESP_NOT_YET_VALID){
        return @"RESP_NOT_YET_VALID";
    }
    return @"UNKNOWN_STATUS";
}


/*
 * simpleValidation
 * Performs a simple OCSP or CRT validation check.
 */
NSDictionary *SimpleValidation(enum Vtk_ValidationMech mech, const char *userCertFile, const char *caCertFile, const char *vaURL, enum Vtk_InputContainer inputType){
    Vtk_Ctxt *ctxt;
    Vtk_Cert *userCert = NULL, *issuerCert = NULL;
    Vtk_Validation *q = NULL;
    Vtk_uint32 ret, status;
    Vtk_ValHdl *hdl = NULL;
    NSMutableDictionary *res = [NSMutableDictionary dictionary];
        
    fprintf(stdout, "\n\nPerforming certificate validation using %s...",
            (mech == VTK_VM_OCSP ? "OCSP" : "CRT"));
    /*
     * Create a toolkit context which will be used for all operations.
     * The context stores configuration information that can be used
     * across validation queries. It comes pre-loaded with
     * the Global Validation Service information and certificates and
     * defaults to using the CRT protocol.
     */
    ctxt = Vtk_CtxtNew();	assert(ctxt != NULL);

    /*
     * Set appropriate validation mechanism for the context.
     * By default, the context is setup with Global Validation
     * Authority Service (GVAS) using the CRT protocol.
     */
    if(!vaURL) vaURL = VTK_GVAS_URL;
    ret = Vtk_CtxtSetDefaultVa(ctxt, vaURL, mech);
    if (ret != VTK_OK){
        showError("Vtk_CtxtSetDefaultVa", ret);
        goto done;
    }

    /*
     * Create a Validation structure.
     * Validation structures encapsulate validation operation to a VA
     * using any of the toolkit supported validation protocols.
     */
    q = Vtk_ValidationNew(ctxt);	assert(q);

    /*
     * Load passed in user and CA certificates.
     */
    userCert = Vtk_CertNew(ctxt);	assert(userCert);
    issuerCert = Vtk_CertNew(ctxt);	assert(issuerCert);

    /*
     * Populate the certificates with information from specified file.
     */
    if(inputType == VA_FILE){
        if ((ret = Vtk_CertLoadFromFile(ctxt, userCert, userCertFile, VTK_DF_BASE64)) != VTK_OK) {
            showError("Vtk_CertLoadFromFile", ret);
            goto done;
        }
        if ((ret = Vtk_CertLoadFromFile(ctxt, issuerCert, caCertFile, VTK_DF_BASE64)) != VTK_OK) {
            showError("Vtk_CertLoadFromFile", ret);
            goto done;
        }
    }else{
        const char *bpuser = [[NSString stringWithCString: userCertFile] cString];
        const char *bpca = [[NSString stringWithCString: caCertFile] cString];
        Vtk_Buffer userBuffer;
        Vtk_Buffer caBuffer;
        VTK_BUF_INIT(&userBuffer);
        VTK_BUF_INIT(&caBuffer);
        userBuffer.dPtr = (Vtk_Byte *)bpuser;
        userBuffer.len = (Vtk_uint32)strlen(userCertFile);
        userBuffer.type = VTK_DF_BASE64;
        caBuffer.dPtr = (Vtk_Byte *)bpca;
        caBuffer.len = (Vtk_uint32) strlen (caCertFile);
        caBuffer.type = VTK_DF_BASE64;
        if ((ret = Vtk_CertInit(ctxt, userCert, &userBuffer)) != VTK_OK) {
            showError("Vtk_CertLoadFromString user", ret);
            goto done;
        }
        if ((ret = Vtk_CertInit(ctxt, issuerCert, &caBuffer)) != VTK_OK) {
            showError("Vtk_CertLoadFromString ca", ret);
            goto done;
        }
    }
    /*
     * Display information about the certificates just loaded.
     */
    //DisplayCert(ctxt, userCert, "User certificate details...");
    //DisplayCert(ctxt, issuerCert, "Issuer certificate details...");

    /*
     * Add the passed in certificate to the validation structure.
     * For each certificate to be validated the toolkit requires the
     * CA certificate along with the certificate to be validated.
     * This operation can be repeated to add other certifcates to be validated.
     * Refer to Vtk_ValidationAddCertRaw for alternative method of suppling the
     * certificate data.
     */
    ret = Vtk_ValidationAddCert(ctxt, q, userCert, issuerCert, &hdl);
    if (ret != VTK_OK){
        showError("Vtk_ValidationAddCert", ret);
        goto done;
    }

    /*
     * Perform the certificate validation.
     * The VA specified in the context will be queried
     * for the certificate(s) status.
     */
    ret = Vtk_ValidationValidate(ctxt, q, &status);

    /*
     * Display validation values.
     */
    if (ret == VTK_OK){
        fprintf(stdout, "\n\nValidation succedded.");
        /*
         * Display the certificate(s) status.
         */
        showStatus("\nOverall certificate(s) status: ", status);
        /*
         * Display detailed information about the performed
         * validtion.
         */
        //displayValidationDetails(ctxt, hdl);
        assignValidationDetails(ctxt, hdl, res);
        [res setObject:@"YES" forKey:@"RESPONSE_STATUS"]; 
    }else{
        fprintf(stdout, "\n\nValidation failed.");
        showError("Vtk_ValidationValidate", ret);
        [res setObject:@"NO" forKey:@"RESPONSE_STATUS"]; 
    }


done:
    // Cleanup memory
    if (q)
        Vtk_ValidationDelete(q);
    if (hdl)
        Vtk_ValHdlDelete(hdl);
    if (userCert)
        Vtk_CertDelete(userCert);
    if (issuerCert)
        Vtk_CertDelete(issuerCert);
    if (ctxt)
        Vtk_CtxtDelete(ctxt);

    return res;
} /* SimpleValidation */


/*
 * Displays validation response details
 *
 * The Vtk_ValHdl is used to obtain detailed validation information
 * for a perticular certificate requsted to be validated.
 * The Vtk_ValRespDetails structure represents the *header* information
 * in a validation response (OCSP, CRT, CRL).
 * The Vtk_ValRespSingleCertDetails structure represents validation
 * information for a specific certificate.
 */
void displayValidationDetails(const Vtk_Ctxt *ctxt, const Vtk_ValHdl *hdl)
{
    Vtk_ValRespDetails *hdr;
    Vtk_ValRespSingleCertDetails *certDetails;
    const char* protocols[] = { "CRT", "OCSP", "CRL" };
    Vtk_uint32 ret, status;
    char *issuer;
    /*
     * Obtain validation details using a validation handle
     * created during Vtk_ValidationAddCert.
     */
    if ((ret = Vtk_ValHdlGetRevStatus(ctxt, hdl, &status, &hdr, &certDetails)) != VTK_OK){
        showError("Vtk_ValidationGetRevStatus", ret);
        return;
    }
    if (hdr){
        /*
         * Print Validation Header information.
         */
        fprintf(stdout, "\nValidation response details...");
        fprintf(stdout, "\n\tProtocol %s,", protocols[hdr->type - 1]);
        fprintf(stdout, "\n\tVersion: %d,\t Issue time: %s", hdr->version, ctime(&hdr->issueTime));
        /*
         * Only one of issuerIdByName or issueridByKey will be set.
         */
        if (hdr->issuerIdByName)
            issuer = (char*)hdr->issuerIdByName->dPtr;
        else
            issuer = (char*)hdr->issuerIdByKey->dPtr;
        fprintf(stdout, "\tIssuer: %s", issuer);
        /*
         * Display additional, protocol specific details.
         */
        if (hdr->type == VTK_VM_CRT){
            fprintf(stdout, "\n\tCRT next update: %s", ctime(&hdr->d.crt->nextUpdate));
            fprintf(stdout, "\tCRT valid untill: %s", ctime(&hdr->d.crt->validUntil));
        }
    }
    /*
     * We can additionally display detailed revocation information
     * for a perticular certificate.
     */
    fprintf(stdout, "\nSingle certificate details...");
    fprintf(stdout, "\n\tCertificate status: 0x%X: ", status);
    showStatus(NULL, status);
    if (certDetails){
        fprintf(stdout, "\n\tThis update: %s", (certDetails->thisUpdate == -1 ? "not specifed\n" : ctime(&certDetails->thisUpdate)));
        fprintf(stdout, "\tNext update: %s", (certDetails->nextUpdate == -1 ? "not specified" : ctime(&certDetails->nextUpdate)));
       /*
        * If this certificate is revoked, there may be additional
        * information.
        */
        if (certDetails->certStatus & VTK_STATUS_REVOKED){
            /* revocation time is optinal parameter -1 means not set */
            if (certDetails->revocationTime != -1)
                fprintf(stdout, "\n\tRevocation time: %s", ctime(&certDetails->revocationTime));

            /* revocation reason is optional field -1 means not set */
            if (certDetails->revocationReason != VTK_REV_STATUS_UNKNOWN)
                fprintf(stdout, "\tRevocation Reason: %d", certDetails->revocationReason);
        }
    }
    fprintf(stdout, "\n\n");
    /*
     * Release memory occupied by structures.
     */
    if (hdr)
        Vtk_ValRespDetailsDelete(hdr);
    if (certDetails)
        Vtk_ValRespSingleCertDetailsDelete(certDetails);
} /* displayValidationDetails */



/*
 * DisplayCert
 *
 * Prints detailed information for the suppied certificate.
 */
void DisplayCert(const Vtk_Ctxt *ctxt, const Vtk_Cert *cert, const char *label){
    Vtk_CertInfo *info;
    Vtk_uint32 ret;
    if (!ctxt || !cert){
        fprintf(stdout, "\n**ERROR** Invalid arguments in DisplayCert.");
        return;
    }
    /*
        * Obtain the X.509 certificate details.
     */
    if ((ret = Vtk_CertGetInfo(ctxt, cert, &info)) != VTK_OK){
        showError("DisplayCert", ret);
        return;
    }
    /*
        * Display certificate details
     */
    fprintf(stdout, "\n\n%s", label);
    fprintf(stdout, "\n\tVersion: %d", info->version);
    fprintf(stdout, "\n\tSerial Num: %s", info->serialNum.dPtr);
    fprintf(stdout, "\n\tIssuer: %s", info->issuer.dPtr);
    fprintf(stdout, "\n\tSubject: %s", info->subject.dPtr);
    fprintf(stdout, "\n\tValid from: %s", ctime(&info->notBefore));
    fprintf(stdout, "\tValid to: %s", ctime(&info->notAfter));
    fprintf(stdout, "\tSignature Algorithm: %s", info->signatureAlg.dPtr);
    fprintf(stdout, "\n\tPublicKey Algorithm: %s", info->publicKeyAlg.dPtr);
    if (info->issuerUniqueId)
        fprintf(stdout, "\n\tIssuer Unique id of len %d set.", info->issuerUniqueId->len);
    if (info->subjectUniqueId)
        fprintf(stdout, "\n\tSubject Unique id of len %d set.", info->subjectUniqueId->len);
    fprintf(stdout, "\n");
    /*
        * Release memory allocated by Vtk_CertGetInfo
     */
    Vtk_CertInfoDelete(info);
} /* DisplayCert */


/*
 * showError
 *
 * Displays a textual representation of the Validator Toolkit error
 */
void showError(const char *funcName, Vtk_uint32 code){
    fprintf(stdout, "\n\n**ERROR** in %s - 0x%X (%s)\n\n",
            funcName, code, Vtk_ErrorToString(code));
} /* showError */


/*
 * showStatus
 *
 * Displays textual representation of the passed in validation
 * status code.
 */
static void showStatus(const char *label, Vtk_uint32 status){
    char ** statusStrings;
    statusStrings = Vtk_StatusToStrings(status);
    /*
     * Ensure Vtk_StatusToStrings worked.
     */
    if (statusStrings){
        char **temp = statusStrings;
        if (label)
            fprintf(stdout, label);
        /*
         * Walk through the array of strings returned, printing each string,
         * until the end of array is reached - NULL entry.
         */
        while (*temp)
            fprintf(stdout, "%s\t", *temp++);
        /*
         * Release status array.
         */
        Vtk_StatusStringsDelete(statusStrings);
    }
    else
        fprintf(stdout,
                "\n\n**ERROR - Vtk_StatusToStrings - no status strings returned\n\n");
} /* showStatus */

static void assignValidationDetails(const Vtk_Ctxt *ctxt, const Vtk_ValHdl *hdl, NSMutableDictionary *res)
{
    Vtk_ValRespDetails *hdr;
    Vtk_ValRespSingleCertDetails *certDetails;
    const char* protocols[] = { "CRT", "OCSP", "CRL" };
    Vtk_uint32 ret, status;
    char *issuer;
    /*
     * Obtain validation details using a validation handle
     * created during Vtk_ValidationAddCert.
     */
    if ((ret = Vtk_ValHdlGetRevStatus(ctxt, hdl, &status, &hdr, &certDetails)) != VTK_OK){
        showError("Vtk_ValidationGetRevStatus", ret);
        return;
    }
    if (hdr){
        /*
         * Print Validation Header information.
         */
        fprintf(stdout, "\nValidation response details...");
        fprintf(stdout, "\n\tProtocol %s,", protocols[hdr->type - 1]);
        fprintf(stdout, "\n\tVersion: %d,\t Issue time: %s", hdr->version, ctime(&hdr->issueTime));
        [res setObject:[NSString stringWithCString: protocols[hdr->type - 1]] forKey:@"PROTOCOL"];
        [res setObject:[NSString stringWithCString: ctime(&hdr->issueTime)] forKey:@"ISSUE_TIME"];
        [res setObject:[NSNumber numberWithInt: hdr->version] forKey:@"VERSION"];
        /*
         * Only one of issuerIdByName or issueridByKey will be set.
         */
        if (hdr->issuerIdByName)
            issuer = (char*)hdr->issuerIdByName->dPtr;
        else
            issuer = (char*)hdr->issuerIdByKey->dPtr;
        fprintf(stdout, "\tIssuer: %s", issuer);
        [res setObject:[NSString stringWithCString: issuer] forKey:@"ISSUER"];
        /*
         * Display additional, protocol specific details.
         */
        if (hdr->type == VTK_VM_CRT){
            fprintf(stdout, "\n\tCRT next update: %s", ctime(&hdr->d.crt->nextUpdate));
            fprintf(stdout, "\tCRT valid untill: %s", ctime(&hdr->d.crt->validUntil));
            [res setObject:[NSString stringWithCString: ctime(&hdr->d.crt->nextUpdate)] forKey:@"NEXT_CRT_UPDATE"];
            [res setObject:[NSString stringWithCString: ctime(&hdr->d.crt->validUntil)] forKey:@"CRT_VALID_UNTIL"];
        }
    }
    /*
     * We can additionally display detailed revocation information
     * for a perticular certificate.
     */
    fprintf(stdout, "\nSingle certificate details...");
    fprintf(stdout, "\n\tCertificate status: 0x%X: ", status);
    [res setObject: statusMessageForStatus(status) forKey:@"CERT_STATUS"];
    showStatus(NULL, status);
    if (certDetails){
        fprintf(stdout, "\n\tThis update: %s", (certDetails->thisUpdate == -1 ? "not specifed\n" : ctime(&certDetails->thisUpdate)));
        fprintf(stdout, "\tNext update: %s", (certDetails->nextUpdate == -1 ? "not specified" : ctime(&certDetails->nextUpdate)));
       /*
        * If this certificate is revoked, there may be additional
        * information.
        */
        if (certDetails->certStatus & VTK_STATUS_REVOKED){
            /* revocation time is optinal parameter -1 means not set */
            if (certDetails->revocationTime != -1){
                fprintf(stdout, "\n\tRevocation time: %s", ctime(&certDetails->revocationTime));
                [res setObject: [NSString stringWithCString: ctime(&certDetails->revocationTime)] forKey:@"REVOCATION_TIME"];
            }
            /* revocation reason is optional field -1 means not set */
            if (certDetails->revocationReason != VTK_REV_STATUS_UNKNOWN){
                fprintf(stdout, "\tRevocation Reason: %d", certDetails->revocationReason);
                [res setObject: [NSNumber numberWithInt:certDetails->revocationReason] forKey:@"REVOCATION_REASON"];
            }
        }
    }
    fprintf(stdout, "\n\n");
    /*
     * Release memory occupied by structures.
     */
    if (hdr)
        Vtk_ValRespDetailsDelete(hdr);
    if (certDetails)
        Vtk_ValRespSingleCertDetailsDelete(certDetails);
} /* displayValidationDetails */
