/*
 * Copyright (c) 2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Portions Copyright (c) 2000 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.1 (the "License").  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON- INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

/*
 *	CGI - the 'simplest' case of an adaptor API.
 *
 *	Headers are passed in as environment variables (char **envp).
 *	Form data is available at stdin.
 *
 *	Response is returned via stdout.
 *
 */
#include "womalloc.h"
#include "MoreURLCUtilities.h"
#include "request.h"
#include "response.h"
#include "errors.h"
#include "log.h"
#include "transaction.h"
#include "listing.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#if !defined(WIN32)
#include <sys/param.h>
#else
#include <io.h>	/* setmode() */
#include <fcntl.h>
#endif

static const char *documentRoot();

#define CGI_SCRIPT_NAME	"SCRIPT_NAME"
#define	CGI_PATH_INFO	"PATH_INFO"
#define	CGI_SERVER_PROTOCOL	"SERVER_PROTOCOL"
#define	CGI_DOCUMENT_ROOT "DOCUMENT_ROOT"
#define WO_CONFIG_URL "WO_CONFIG_URL"
#define WO_XYZZY_USERNAME "WO_XYZZY_USERNAME"
#define WO_XYZZY_PASSWORD "WO_XYZZY_PASSWORD"

static unsigned int freeValueNeeded=0;

/*
 * BEGIN Support for getting the client's certificate.
 */
char *make_cert_one_line(char *value) {
    char *returnValue = (char *)NULL;
    int i;
    int j;

    if (value) {
        //
        // Copy everything except newlines.
        // Really, we should copy everything that is in the base64
        // encoding alphabet except for newlines, spaces, and tabs.
        // But, this will do for now...
        //
        returnValue = strdup(value);
        freeValueNeeded=1;
        
        for (j=0, i = 0; i < strlen(value);i++) {
            if (value[i] != '\n') {
                returnValue[j]= value[i];
                j++;
            }
        }
        returnValue[j] = '\0';  // and NULL-terminate
    }

    return returnValue;
}
/*
 * END Support for getting the client's certificate.
 */



/*
 *	the CGI1.1 spec says:
 *
 *	"For Unix compatible operating systems, the following are defined:
 *	...
 *	Character set
 *		The US-ASCII character set is used for the definition of
 *		environment variable names and header field names; the newline
 *		(NL) sequence is LF; servers SHOULD also accept CR LF as a
 *		newline.
 *	..."
 */
#define	CRLF	"\r\n"

#define	HDRBUFSZ	6000
/*
 *	send response to server
 */
static void sendResponse(HTTPResponse *resp)
{
    char *resphdrs;
    char buffer[HDRBUFSZ];
#ifndef PROFILE
    fprintf(stdout,"Status: %d %s" CRLF,resp->status,resp->statusMsg);
#endif
    /* Laurent R-: Radar bug 2243689 - content-length should not be sent
     * not twice, not once :-)
     */
#ifdef WIN32
    st_removeKey(resp->headers, CONTENT_LENGTH);
#endif

    resphdrs = resp_packageHeaders(resp, buffer, HDRBUFSZ);
#ifndef PROFILE
    fputs(resphdrs,stdout);
#endif
    if (resphdrs != buffer)	{	/* did we use our static buffer? */
        free(resphdrs);
    }
#ifndef PROFILE
    fputs(CRLF,stdout);		/* RADAR 2243694: follow w3c spec. */
    fputs(CRLF,stdout);
    fflush(stdout);			/* #2312916 Fix problem with WebObjects CGI not terminating if socket closes early */
#endif

#ifndef PROFILE
    if (resp->content_length) {
        fwrite(resp->content,sizeof(char),resp->content_length,stdout);
        fflush(stdout);
    }
#endif
    return;		
}

static void die_resp(HTTPResponse *resp)
{
    sendResponse(resp);
    resp_free(resp);
    exit(0);
}

static void die(const char *msg)
{
    HTTPResponse *resp = resp_errorResponse(msg);
    die_resp(resp);
}

#ifdef	PROFILE
int doit(int argc, char *argv[], char **envp);	/* forward */

int main(int argc, char *argv[], char **envp) {
    int i;
    for (i=0; i < 50000; i++)
        doit(argc, argv, envp);
    return 0;
}
int doit(int argc, char *argv[], char **envp) {
#else
/*
 *	the request handler...
 */
int main(int argc, char *argv[], char **envp) {
#endif
    HTTPRequest *req;
    HTTPResponse *resp = NULL;
    WOURLComponents wc = WOURLComponents_Initializer;
    const char *qs;
    unsigned int qs_len;
    char *url;
    const char *script_name, *path_info, *config_url, *username, *password;
    const char *reqerr;
    WOURLError urlerr;
    strtbl *options = NULL;

    /* Laurent R-: fixed radar bug 2246729 */
#ifdef WIN32
    _setmode(_fileno(stdout), _O_BINARY);
    _setmode(_fileno(stdin), _O_BINARY);
#endif

    script_name = getenv(CGI_SCRIPT_NAME);
    path_info = getenv(CGI_PATH_INFO);

    if (script_name == NULL)
        die(INV_SCRIPT);
    else if (path_info == NULL) {
        path_info = "/";
    }
    /* #2350653 Provide a hook via an environment variable to define the config URL */
    config_url = getenv(WO_CONFIG_URL);
    if (!config_url) {
        /* Flat file URL */
        /* config_url = "file:///Local/Library/WebObjects/Configuration/WOConfig.xml"; */
        /* Local wotaskd */
        /* config_url = "http://localhost:1085"; */
        /* Multicast URL */
        config_url = CONFIG_URL; /* Actually "webobjects://239.128.14.2:1085"; */
    }
    options = st_new(8);
    st_addStatic(options, WOCONFIG, config_url);

    /* 
     * If your webserver is configured to pass these environment variables, we use them to
     * protect xyzzy output.
     */
    username = getenv(WO_XYZZY_USERNAME);
    if (username && strlen(username) != 0) {
        st_addStatic(options, WOUSERNAME, username);
        password = getenv(WO_XYZZY_PASSWORD);
        if(password && strlen(password) != 0) {
            st_addStatic(options, WOPASSWORD, password);
        }
    }
    
    /*
     * SECURITY ALERT
     *
     * To disable xyzzy, uncomment the next line. 
     * st_addStatic(options, WOUSERNAME, "disabled");
     *
     * To specify an xyzzy username and password, uncomment the next two lines.
     * st_addStatic(options, WOUSERNAME, "joe");
     * st_addStatic(options, WOPASSWORD, "secret");
     *
     */

    init_adaptor(options);

    /*
     *	extract WebObjects application name from URI
     */

    /*
     * #2411402 We need to know this to decide the size of url up front
     */
    qs = getenv("QUERY_STRING");

    if (qs) {
        qs_len = strlen(qs);
    } else {
        qs_len = 0;
    }

    if (qs_len > 0) {
        /* Make room for query string and "?" */
        url = WOMALLOC(strlen(path_info) + strlen(script_name) + 1 + qs_len + 2);
    } else {
        url = WOMALLOC(strlen(path_info) + strlen(script_name) + 1);
    }
    strcpy(url, script_name);
    strcat(url, path_info);
    WOLog(WO_INFO,"<CGI> new request: %s",url);

    urlerr = WOParseApplicationName(&wc, url);
    if (urlerr != WOURLOK) {
        const char *_urlerr;
        _urlerr = WOURLstrerror(urlerr);
        WOLog(WO_INFO,"URL Parsing Error: %s", _urlerr);
     
        if ((urlerr == WOURLInvalidApplicationName) && ac_shouldList()) {
            resp = WOConfigFileListing(&wc);
            die_resp(resp);
        }
        
        die(_urlerr);
    }

    /*
     *	build the request...
     */
    req = req_new( getenv("REQUEST_METHOD"), NULL);

    /*
     *	validate the method
     */
    reqerr = req_validateMethod(req);
    if (reqerr) {
        die(reqerr);
    }

    /*
     *	copy the headers.  This looks wierd... all we're doing is copying
     *	*every* environment variable into our headers.  It may be beyond
     *	the spec, but more information probably won't hurt.
     */
    while (envp && *envp) {
        char *key, *value;
        /* we are leaking here, but this is only cgi, so it will go away. */
        /* copy env. line. */
        key = WOSTRDUP(*envp);

        for (value = key; *value && !isspace(*value) && (*value != '='); value++) {}
        if (*value) {
            *value++ = '\0';	/* null terminate 'key' */
        }
        while (*value && (isspace(*value) || (*value == '='))) {
            value++;
        }

        WOLog(WO_INFO,"<CGI> ENV VAR(%s) = %s", key,value);

        /* BEGIN Support for getting the client's certificate. */
        if (strcmp((const char *)key, "CLIENT_CERT") == 0) {
          value = make_cert_one_line(value);
          WOLog(WO_INFO,"<CGI> PASSING CLIENT_CERT = %s", value);
        }
        /*  END Support for getting the client's certificate  */
        
        if (key && *key && value && *value) {
            req_addHeader(req, key, value);
        }

        /*  BEGIN Support for getting the client's certificate  */
        if (freeValueNeeded ) {
            free(value);
            freeValueNeeded=0;
            WOLog(WO_DBG,"<CGI> freeing cert value strdup memory");
        }
        /*  END Support for getting the client's certificate  */
        
        envp++;			/* next env variable */
    }

    /*
     *	get form data if any
     */
    if ((req->method == HTTP_POST_METHOD) && (req->content_length > 0)) {
        char *buffer = WOMALLOC(req->content_length);
        int n = fread(buffer, 1, req->content_length, stdin);
        if (n != req->content_length) {
            const char *errmsg;
            int err = ferror(stdin);
            if (err) {
                WOLog(WO_ERR,"Error getting FORM data: %s (%d)", strerror(errno), err);
                errmsg = NO_FORM_DATA;
            } else {
                WOLog(WO_ERR,"Error getting FORM data: incorrect content-length (expected: %d read: %d)%s", req->content_length, n, strerror(errno));
                errmsg = INV_FORM_DATA;
            }
            WOFREE(buffer);
            die(errmsg);
        } else {
            req->content = buffer;
            /* Radar 2396884 -- disabled as content not always printable */
            /*    WOLog(WO_DBG,"<CGI> new request with Content: %s", buffer);  */
        }
    }  else if (req->method == HTTP_GET_METHOD) {
 
        /* Radar 2398271 Don't add ? if the query string is empty. */
        if (qs_len > 0) {
            /* Add the query string to the full URL - useful for debugging */
            /* #2347762 Get enough space for this string concatenation */
            /* #2411402 Can't use realloc as wc components are pointers into this url */
            
            strcat(url, "?");
            strcat(url, qs);
            wc.queryString.start = qs;
            wc.queryString.length = qs_len;
            WOLog(WO_INFO,"<CGI> new request with Query String: %s", qs);
        }
    }

    /*
     *	message the application & collect the response
     */
    resp = tr_handleRequest(req, url, &wc, getenv(CGI_SERVER_PROTOCOL), documentRoot());

    if (resp != NULL) {
        sendResponse(resp);
        resp_free(resp);		/* dump the response */
    }

#if defined(FINDLEAKS)
    WOFREE(url);				/* just for neatness */
    st_free(options);

    showleaks();
#endif
    return 0;
}


const char *documentRoot() {
    static char path[MAXPATHLEN+1] = "";

    if (path[0] == '\0') {
#ifdef WIN32
        WOReadKeyFromConfiguration(CGI_DOCUMENT_ROOT, path, MAXPATHLEN);
#else
        const char *doc_root;
        /* Apache provides this as an environment variable straight */
        if ((doc_root = getenv(CGI_DOCUMENT_ROOT)) != NULL) {
            strncpy(path, doc_root, MAXPATHLEN);
        } else {
            const char *path_trans, *path_info;
            path_trans = getenv("PATH_TRANSLATED");
            path_info = getenv("PATH_INFO");

            if (path_trans && path_info) {
                char *e = strstr(path_trans,path_info);
                if (e) {
                    strncpy(path,path_trans,e-path_trans);
                }
            }
        }
#endif
    }
    if (path[0] != '\0')
        return path;
    else {
        WOLog(WO_ERR,"Can't find document root in CGI variables");
        return "/usr/local/apache/htdocs";		/* this is bad.... */
    }
}
 

